from pydantic import BaseModel

from aporacle import conf


class EventData(BaseModel):
    chain: str = conf.chain
    reward_epoch: int = 0
    event: str
    block_number: int
    timestamp: str


class VotingRound(EventData):
    voting_round: int
    voting_round_start: int


#
# from oracle_chain_events.chain.oracle.contracts.feed.tso_watcher import TsoWatcher
# from oracle_chain_events.chain.oracle.contracts.tso_registry.tso_registry_watcher import TsoRegistryWatcher
# from oracle_chain_events.chain.watcher.contract_event_logs import ContractEventLogger
# from oracle_chain_events.chain.watcher.executor import NewBlocksWatcher
#
#
# __all__ = [
#     NewBlocksWatcher,
#     ContractEventLogger,
#     # PriceSubmitterWatcher,
#     TsoWatcher,
#     TsoRegistryWatcher,
# ]



from decimal import Decimal
from typing import Dict

from komoutils.core import KomoBase


class WalletBase(KomoBase):
    @property
    def address(self) -> str:
        raise NotImplementedError

    def get_balance(self, asset_name: str) -> Decimal:
        raise NotImplementedError

    def get_raw_balance(self, asset_name: str) -> int:
        raise NotImplementedError

    def get_all_balances(self) -> Dict[str, Decimal]:
        raise NotImplementedError

    def send(self, address: str, asset_name: str, amount: Decimal) -> str:
        raise NotImplementedError

    def to_nominal(self, asset_name: str, raw_amount: int) -> Decimal:
        raise NotImplementedError

    def to_raw(self, asset_name: str, nominal_amount: Decimal) -> int:
        raise NotImplementedError


import asyncio
from collections import OrderedDict
import cytoolz
import functools
from hexbytes import HexBytes
from eth_bloom import BloomFilter
import logging
from typing import (
    Dict,
    List,
    Optional,
    Set
)

from komoutils.core import KomoBase, safe_gather
from web3 import Web3
from web3.datastructures import AttributeDict
from web3._utils.contracts import find_matching_event_abi
from web3._utils.events import get_event_data
from web3._utils.filters import construct_event_filter_params
from eth_abi.codec import (
    ABICodec,
)
from eth_abi.registry import registry

from aporacle.core.utils.async_call_scheduler import AsyncCallScheduler

DEFAULT_WINDOW_SIZE = 100


class ContractEventLogger(KomoBase):

    def __init__(self,
                 w3: Web3,
                 address: str,
                 contract_abi: List[Dict[str, any]],
                 block_events_window_size: Optional[int] = DEFAULT_WINDOW_SIZE):

        super().__init__()
        self._w3: Web3 = w3
        self._ev_loop: asyncio.BaseEventLoop = asyncio.get_event_loop()
        self._block_events_window_size = block_events_window_size
        self._address: str = address
        self._contract_abi: List[Dict[str, any]] = contract_abi
        self._event_abi_map: Dict[str, Dict[str, any]] = {}
        self._event_cache: Set[HexBytes] = set()
        self._block_events: OrderedDict = OrderedDict()

    @property
    def address(self) -> str:
        return self._address

    @property
    def contract_abi(self) -> List[Dict[str, any]]:
        return self._contract_abi

    async def get_new_entries_from_logs(self,
                                        event_name: str,
                                        blocks: List[AttributeDict]) -> List[tuple[AttributeDict, List[str]]]:
        event_abi: Dict[str, any] = self._event_abi_map.get(event_name, None)
        if event_abi is None:
            event_abi = find_matching_event_abi(self._contract_abi, event_name=event_name)
            self._event_abi_map[event_name] = event_abi

        _, event_filter_params = construct_event_filter_params(event_abi,
                                                               contract_address=Web3.to_checksum_address(self._address),
                                                               abi_codec=ABICodec(registry))

        tasks = []
        for block in blocks:
            block_bloom_filter = BloomFilter(int.from_bytes(block["logsBloom"], byteorder='big'))
            check_block = True
            for topic in event_filter_params["topics"]:
                if not bytes.fromhex(topic.lstrip("0x")) in block_bloom_filter:
                    check_block = False
                    break
            if check_block:
                event_filter_params["blockHash"] = block["hash"].hex()
                tasks.append(self._get_logs(event_filter_params))

        # print(f"Tasks count {len(tasks)}")

        new_entries = []
        if len(tasks) > 0:
            raw_logs = await safe_gather(*tasks, return_exceptions=True)
            logs: List[any] = list(cytoolz.concat(raw_logs))
            for log in logs:
                event_data: AttributeDict = get_event_data(ABICodec(registry), event_abi, log)
                # event_data["topics"] = log['topics'][0]
                event_data_block_number: int = event_data["blockNumber"]
                event_data_tx_hash: HexBytes = event_data["transactionHash"]
                if event_data_tx_hash not in self._event_cache:
                    if event_data_block_number not in self._block_events:
                        self._block_events[event_data_block_number] = [event_data_tx_hash]
                    else:
                        self._block_events[event_data_block_number].append(event_data_tx_hash)
                    self._event_cache.add(event_data_tx_hash)
                    new_entries.append((event_data, log['topics']))
                else:
                    self.log_with_clock(log_level=logging.DEBUG,
                                        msg=f"Duplicate event transaction hash found - '{event_data_tx_hash.hex()}'.")

            while len(self._block_events) > self._block_events_window_size:
                tx_hashes: List[HexBytes] = self._block_events.popitem(last=False)[1]
                for tx_hash in tx_hashes:
                    self._event_cache.remove(tx_hash)

        # print(f"New Entries {new_entries}")
        return new_entries

    async def _get_logs(self,
                        event_filter_params: Dict[str, any],
                        max_tries: Optional[int] = 30) -> List[Dict[str, any]]:
        async_scheduler: AsyncCallScheduler = AsyncCallScheduler.shared_instance()
        count: int = 0
        logs = []
        while True:
            try:
                count += 1
                if count > max_tries:
                    self.log_with_clock(log_level=logging.DEBUG,
                                        msg=f"Error fetching logs from block with filters: '{event_filter_params}'.")
                    break
                logs = await async_scheduler.call_async(
                    functools.partial(self._w3.eth.get_logs, event_filter_params)
                )
                break
            except asyncio.CancelledError:
                raise
            except Exception:
                self.log_with_clock(log_level=logging.DEBUG,
                                    msg=f"Block not found with filters: '{event_filter_params}'. Retrying...")
                await asyncio.sleep(0.5)
        return logs



import inspect
import requests

from aporacle.connectors.rpc.flarepy.api.apimeta import api_endpoints

host = "flare-api.flare.network"
# jsonrpc_port = "9650"
protocol = "https"
# jsonrpc_url = "{}://{}:{}".format(jsonrpc_prot, jsonrpc_host, jsonrpc_port)
url = "{}://{}".format(protocol, host)

urls = {k: ("{}" + v).format(url) for k, v in api_endpoints.items()}

# print(f"URLS {urls}")


def chain_call(url, method, params=None):
    payload = {
        "method": method,
        "params": params,
        "jsonrpc": "2.0",
        "id": 1
    }
    print(url)
    print(payload)

    # return
    # url = 'https://flare-api.flare.network/ext/bc/P'
    # payload = {"jsonrpc": "2.0", "method": "platform.getCurrentValidators", "params": {}, "id": 1}

    response = requests.post(url, json=payload).json()

    if "error" in response:
        print(response["error"]["message"])
        return

    return response["result"]


def chain_post(url, params=None):
    return requests.post(url, params)


def make_caller(url):
    def f(method, data):
        return chain_call(url, method, data)

    return f


def make_poster(url):
    def f(data):
        return chain_post(url, data)

    return f


def get_caller():
    """
    Uses the source caller's module name to determine
    which caller to return.
    Will fail if source is __main__. Do not call module
    functions directly from command line.
    """
    src_caller = inspect.stack()[1]
    module_name = str(inspect.getmodule(src_caller[0]).__name__).split('.')[-1]
    print(str(inspect.getmodule(src_caller[0]).__name__))
    return make_caller(urls[module_name])


def get_poster():
    src_caller = inspect.stack()[1]
    module_name = inspect.getmodule(src_caller[0])
    return make_poster(urls[module_name])



from typing import Dict

from web3 import Web3
from web3.middleware import geth_poa_middleware

from aporacle import conf
from aporacle.connectors.evm_chain import FlareChain
from aporacle.connectors.oracle.contracts.flare_system_manager.flare_systems_manager_contract import \
    FlareSystemsManagerContract
from aporacle.connectors.oracle.contracts.submission.submission_contract import SubmissionContract
from aporacle.connectors.oracle.contracts.tso_contract_base import TsoContractBase
from aporacle.connectors.wallet_base import WalletBase


class Web3Wallet(WalletBase):
    BACKEND_SELECTION_INTERVAL = 15.0
    WALLET_EVENT_DEDUP_WINDOW_SIZE = 1024

    def __init__(self,
                 rpc_url: list,
                 chain: FlareChain = FlareChain.SONGBIRD):
        super().__init__()

        self._w3: Web3 = Web3(Web3.HTTPProvider(rpc_url[0]))
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        self._chain: FlareChain = chain

        # Initialize price provider contract data structures.
        self._flare_contracts = {
            "flare_systems_manager": FlareSystemsManagerContract(w3=self._w3, address=conf.flare_systems_manager_address),
            "contract": SubmissionContract(w3=self._w3, address=conf.submission_address),
        }

    @property
    def w3(self) -> Web3:
        return self._w3

    @property
    def chain(self) -> FlareChain:
        return self._chain

    @property
    def reward_epoch(self) -> int:
        return self._flare_contracts["contract"].reward_epoch

    @property
    def current_voting_round(self) -> int:
        return self._flare_contracts["contract"].current_voting_round

    @property
    def flare_contracts(self) -> Dict[str, TsoContractBase]:
        return self._flare_contracts.copy()

    def get_asset_price(self, asset: str) -> float:
        return self._flare_contracts["contract"].get_asset_price(asset)




import asyncio
import logging
from decimal import Decimal
from typing import (
    Dict,
    Optional
)

from komoutils.core import KomoBase, safe_ensure_future, PublishQueue
from web3 import Web3
from web3.contract import Contract
from web3.middleware import geth_poa_middleware

from aporacle import conf
from aporacle.connectors.evm_chain import FlareChain
from aporacle.connectors.oracle.contracts.submission.submission_contract import SubmissionContract
from aporacle.connectors.oracle.contracts.submission.submission_watcher import SubmissionWatcher
from aporacle.connectors.oracle.contracts.tso_contract_base import TsoContractBase
from aporacle.connectors.watcher.websocket_watcher import WSNewBlocksWatcher
from aporacle.network.network_base import NetworkStatus

s_decimal_0 = Decimal(0)


class Web3WalletBackend(KomoBase):
    def __init__(self,
                 rpc_url: str,
                 websocket_url: str,
                 event_output_queue: Optional[asyncio.Queue],
                 chain: FlareChain = FlareChain.SONGBIRD):
        super().__init__()

        self._chain = chain
        # Initialize Web3, accounts and contracts.
        self._w3: Web3 = Web3(Web3.HTTPProvider(rpc_url))
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        self._chain: FlareChain = chain
        self._websocket_url = websocket_url
        self._event_output_queue = event_output_queue

        # Contracts
        self._flare_v2_contracts: Dict[str, TsoContractBase] = {}

        # Blockchain data
        self._local_nonce: int = -1

        # Watchers
        self._new_blocks_watcher: Optional[WSNewBlocksWatcher] = None
        self._submission_watcher: Optional[SubmissionWatcher] = None

        self._blocks_publishing_queue: PublishQueue = PublishQueue()

        # Network
        self._check_network_task: Optional[asyncio.Task] = None
        self._network_status: NetworkStatus = NetworkStatus.STOPPED

        self.set_chain_contracts()

    @property
    def block_number(self) -> int:
        return self._new_blocks_watcher.block_number if self._new_blocks_watcher is not None else -1

    @property
    def chain(self) -> FlareChain:
        return self._chain

    @property
    def flare_v2_contracts(self) -> Dict[str, TsoContractBase]:
        return self._flare_v2_contracts

    @property
    def started(self) -> bool:
        return self._check_network_task is not None

    @property
    def network_status(self) -> NetworkStatus:
        return self._network_status

    @property
    def contracts(self) -> Dict[str, Contract]:
        return {name: contract.contract for name, contract
                in self.flare_v2_contracts.items()}

    def set_chain_contracts(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Updating chain contracts. ")

        # Initialize price provider contract data structures.
        self._flare_v2_contracts = {
            # "flare_systems_manager": FlareSystemsManagerContract(w3=self._w3,
            #                                                      address=conf.flare_systems_manager_address),
            "contract": SubmissionContract(w3=self._w3, address=conf.submission_address),
        }

    def start(self):
        if self.started:
            self.stop()

        self._check_network_task = safe_ensure_future(self._check_network_loop())
        self._network_status = NetworkStatus.NOT_CONNECTED

    def stop(self):
        if self._check_network_task is not None:
            self._check_network_task.cancel()
            self._check_network_task = None
        safe_ensure_future(self.stop_network())
        self._network_status = NetworkStatus.STOPPED

    async def start_network(self):
        self.log_with_clock(log_level=logging.INFO, msg=f"Starting back end.")
        # Create event watchers.
        self._new_blocks_watcher = WSNewBlocksWatcher(self._w3, self._websocket_url)
        self._submission_watcher = SubmissionWatcher(
            w3=self._w3,
            contract=self._flare_v2_contracts["contract"],
            blocks_watcher=self._new_blocks_watcher,
            event_output_queue=self._event_output_queue
        )

        await self._new_blocks_watcher.start_network()
        await self._submission_watcher.start_network()

    async def stop_network(self):
        # Stop the event watchers.
        if self._new_blocks_watcher is not None:
            await self._new_blocks_watcher.stop_network()
            self._new_blocks_watcher = None
            del self._new_blocks_watcher
        if self._submission_watcher is not None:
            await self._submission_watcher.stop_network()

    async def check_network(self) -> NetworkStatus:
        # Assume connected if received new blocks in last 2 minutes
        # if time.time() - self._last_timestamp_received_blocks < 60 * 2:
        return NetworkStatus.CONNECTED

        try:
            await self._update_gas_price()
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED

    async def _check_network_loop(self):
        while True:
            try:
                new_status = await asyncio.wait_for(self.check_network(), timeout=10.0)
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                new_status = NetworkStatus.NOT_CONNECTED
            except Exception:
                self.log_with_clock(log_level=logging.ERROR,
                                    msg="Unexpected error while checking for network status.", exc_info=True)
                new_status = NetworkStatus.NOT_CONNECTED

            self._network_status = new_status
            await asyncio.sleep(5.0)



#!/usr/bin/env python
import asyncio
import logging
from decimal import Decimal
from typing import (
    List,
    Optional
)

import pendulum
from komoutils.core import safe_ensure_future
from web3 import Web3
from web3.datastructures import AttributeDict

from aporacle.connectors.oracle.contracts.tso_contract_base import TsoContractBase
from aporacle.connectors.watcher.base_watcher import BaseWatcher
from aporacle.connectors.watcher.contract_event_logs import ContractEventLogger
from aporacle.connectors.watcher.websocket_watcher import WSNewBlocksWatcher

s_decimal_0 = Decimal(0)

NEW_VOTING_ROUND_INITIATED_EVENT_NAME = 'NewVotingRoundInitiated'


class FtsoFeedPublisherWatcher(BaseWatcher):
    def __init__(self,
                 w3: Web3,
                 contract: TsoContractBase,
                 blocks_watcher: WSNewBlocksWatcher,
                 event_output_queue: Optional[asyncio.Queue],
                 ):
        super().__init__(w3)
        self._blocks_watcher: WSNewBlocksWatcher = blocks_watcher
        self._contract = contract.contract
        self._contract_event_logger = ContractEventLogger(w3, contract.address, contract.abi)
        self._event_output_queue = event_output_queue

        self._new_blocks_queue: asyncio.Queue = self._blocks_watcher.blocks_publishing_queue.register()
        self._poll_logs_task: Optional[asyncio.Task] = None

    async def start_network(self):
        self._poll_logs_task = asyncio.ensure_future(self.poll_logs_loop())

    async def stop_network(self):
        if self._poll_logs_task is not None:
            self._poll_logs_task.cancel()
            self._poll_logs_task = None

    def did_receive_new_blocks(self, new_blocks: List[AttributeDict]):
        self._new_blocks_queue.put_nowait(new_blocks)

    async def poll_logs_loop(self):
        while True:
            try:
                new_blocks: List[AttributeDict] = await self._new_blocks_queue.get()
                # Process Events
                new_voting_entries = await self._contract_event_logger.get_new_entries_from_logs(
                    NEW_VOTING_ROUND_INITIATED_EVENT_NAME,
                    new_blocks
                )
                for new_voting_entry in new_voting_entries:
                    await self._handle_event_data(new_voting_entry)

            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                continue
            except Exception as e:
                print(e)

    async def _handle_event_data(self, event_data: AttributeDict):
        # print('--------------------------------------------------')
        # print(event_data)
        # print('--------------------------------------------------')
        event_type: str = event_data["event"]
        # timestamp: float = float(await self._blocks_watcher.get_timestamp_for_block(event_data["blockHash"]))
        # tx_hash: str = event_data["transactionHash"].hex()
        if event_type == NEW_VOTING_ROUND_INITIATED_EVENT_NAME:
            self.handle_new_voting_round_event(event_data)

    def handle_new_voting_round_event(self, event_data: AttributeDict):
        stamp = pendulum.now("UTC").int_timestamp

        # vr: VotingRound = VotingRound(
        #     event='new_voting_round_initiated',
        #     voting_round=stamp,
        #     voting_round_start=stamp,
        #     voting_round_end=stamp + conf.chain_voting_round_duration,
        #     block_number=event_data["blockNumber"],
        #     timestamp=pendulum.now("UTC").to_iso8601_string()
        # )
        # safe_ensure_future(self._event_output_queue.put(vr.model_dump()))
        # self.log_with_clock(log_level=logging.INFO, msg=f"New voting round issued at block - {vr.block_number}. ")



from aporacle.connectors.rpc.flarepy.api import flarerpc

caller = flarerpc.get_caller()


def get_current_validators(subnet_id=None, node_ids=None):
    data = {}

    if subnet_id:
        data["subnetID"] = subnet_id

    if node_ids:
        data["nodeIDs"] = node_ids

    return caller("platform.getCurrentValidators", data)
